from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey, EllipticCurvePublicKey
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import hmac
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import os

n = 30

def generate_private_key(encryption_alg: ec.EllipticCurve) -> EllipticCurvePrivateKey:
    return ec.generate_private_key(encryption_alg)


def generate_public_key(private_key: EllipticCurvePrivateKey) -> EllipticCurvePublicKey:
    return private_key.public_key()

def print_keys_info(private_key):
    public_key = generate_public_key(private_key)
    print(n*'*')
    print("PRIVATE KEY INFO")
    print(f"k: {private_key.private_numbers().private_value}")
    print(n * '*')
    print("PUBLIC KEY INFO")
    print(f"Curve: {public_key.public_numbers().curve.name}")
    print(f"X: {public_key.public_numbers().x}")
    print(f"Y: {public_key.public_numbers().y}")
    print(n * '*')

def print_private_key_info(private_key):
    print(n*'*')
    print("PRIVATE KEY INFO")
    print(f"k: {private_key.private_numbers().private_value}")
    print(n * '*')


def print_public_key_info(public_key):
    print(n * '*')
    print("PUBLIC KEY INFO")
    print(f"Curve: {public_key.public_numbers().curve.name}")
    print(f"X: {public_key.public_numbers().x}")
    print(f"Y: {public_key.public_numbers().y}")
    print(n * '*')

def generate_signature(data, key: EllipticCurvePrivateKey, hash_alg: HashAlgorithm) -> bytes:
    return key.sign(data, ec.ECDSA(hash_alg))


def verify_signature(key: EllipticCurvePublicKey, signature: bytes, data: bytes, hash_alg: HashAlgorithm) -> bool:
    try:
        key.verify(signature, data, ec.ECDSA(hash_alg))
    except InvalidSignature:
        return False
    return True


def store_keys(private_key):
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )

    public_key = generate_public_key(private_key)
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    with open("private_key.pem", 'wb') as f:
        f.write(private_key_pem)
        f.close()

    with open("public_key.pem", 'wb') as f:
        f.write(public_key_pem)
        f.close()


def load_private_key(private_key_path) -> EllipticCurvePrivateKey:
    with open(private_key_path, "rb") as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=None,
            backend=default_backend()
        )
        print(private_key)
        key_file.close()
    return private_key


def load_public_key(public_key_path) -> EllipticCurvePublicKey:
    with open(public_key_path, "rb") as key_file:
        public_key = serialization.load_pem_public_key(
            key_file.read(),
            backend=default_backend()
        )
        key_file.close()

    return public_key


def pem_to_public_key(public_pem) -> EllipticCurvePublicKey:
    return serialization.load_pem_public_key(
            public_pem,
            backend=default_backend()
        )


def pem_to_private_key(private_pem) -> EllipticCurvePrivateKey:
    return serialization.load_pem_private_key(
            private_pem,
            password=None,
            backend=default_backend()
        )


def ecies_encrypt(public_key, plaintext, encryption_alg, hash_alg):
    # Load the public key from PEM format
    #public_key = serialization.load_pem_public_key(public_key_pem)
    # Generate an ephemeral (temporary) private key
    ephemeral_private_key = ec.generate_private_key(encryption_alg)

    # Derive a shared secret using ECDH (Elliptic Curve Diffie-Hellman)
    shared_secret = ephemeral_private_key.exchange(ec.ECDH(), public_key)
    # print(f"shared_secret: {shared_secret}")

    # Derive encryption and MAC keys using a new KDF (Key Derivation Function) instance
    # print(f"HASHALG: {hash_alg}")
    encryption_key = PBKDF2HMAC(
        algorithm=eval(hash_alg),
        # salt=os.urandom(16),
        salt=b'',
        iterations=100000,
        length=32  # Specify the key length (32 bytes for AES-256)
    ).derive(shared_secret)

    mac_key = PBKDF2HMAC(
        algorithm=eval(hash_alg),
        # salt=os.urandom(16),
        salt=b'',
        iterations=100000,
        length=32  # Specify the key length (32 bytes for AES-256)
    ).derive(shared_secret)
    # print(f"mac_key: {mac_key}")

    # Encrypt the plaintext using AESF
    iv = os.urandom(16)  # Generate a random initialization vector (IV)
    print(f"iv: {iv}")
    print(f"Len iv: {len(iv)}")

    cipher = Cipher(algorithms.AES(encryption_key), modes.CFB(iv))

    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(plaintext.encode('utf-8')) + encryptor.finalize()
    print(f"ciphertext: {ciphertext}")
    print(f"Len ciphertext: {len(ciphertext)}")
    # Create the ciphertext
    ciphertext = iv + ciphertext
    print(f"iv + ciphertext: {ciphertext}")
    print(f"Len iv + ciphertext: {len(ciphertext)}")

    # Create a MAC (Message Authentication Code) of the ciphertext
    mac = hmac.HMAC(mac_key, eval(hash_alg))
    mac.update(ciphertext)
    mac_fin = mac.finalize()
    print(f"mac_fin: {mac_fin}")
    print(f"len mac_fin: {len(mac_fin)}")

    # Serialize the ephemeral public key to PEM format
    ephemeral_public_pem = ephemeral_private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    print(f"ephemeral_public_pem + ciphertext + mac_fin: {ephemeral_public_pem + ciphertext + mac_fin}")
    print(f"Len ephemeral_public_pem: {len(ephemeral_public_pem)}")
    print(f"Len ephemeral_public_pem + ciphertext + mac_fin: {len(ephemeral_public_pem) + len(ciphertext) + len(mac_fin)}")
    # Return the ciphertext, MAC, and ephemeral public key
    return ephemeral_public_pem + ciphertext + mac_fin

def ecies_decrypt(private_key, encrypted_message, hash_alg):
    print(30*"*")
    # print(f"private_key_pem: {private_key_pem}")
    print(f"encrypted_message: {encrypted_message}")

    # Load the private key from PEM format
    #private_key = serialization.load_pem_private_key(private_key_pem, password=None)
    # Extract the ephemeral public key, ciphertext, and MAC from the input
    hash_size = eval(hash_alg).digest_size
    split_bytes=b'\n-----END PUBLIC KEY-----\n'

    ephemeral_public_pem, ciphertext = encrypted_message.split(split_bytes)  #encrypted_message[:-(hash_size + 37)]  # The ephemeral public key is 65 bytes for SECP256R1
    ephemeral_public_pem += split_bytes

    received_mac = ciphertext[-hash_size:]  # The last 32 bytes are the MAC
    ciphertext = ciphertext[:-hash_size]  # Exclude the ephemeral public key and MAC

    # Deserialize the ephemeral public key
    ephemeral_public_key = serialization.load_pem_public_key(ephemeral_public_pem)

    # Derive a shared secret using ECDH
    shared_secret = private_key.exchange(ec.ECDH(), ephemeral_public_key)
    print(f"shared_secret: {shared_secret}")
    # Derive encryption and MAC keys using a new KDF (Key Derivation Function) instance
    encryption_key = PBKDF2HMAC(
        algorithm=eval(hash_alg),
        # salt=os.urandom(16),
        salt=b'',
        iterations=100000,
        length=32  # Specify the key length (32 bytes for AES-256)
        #length=16
    ).derive(shared_secret)
    # print(f"encryption_key: {encryption_key}")

    mac_key = PBKDF2HMAC(
        algorithm=eval(hash_alg),
        # salt=os.urandom(16),
        salt=b'',
        iterations=100000,
        length=32  # Specify the key length (32 bytes for AES-256)
        #length=16
    ).derive(shared_secret)


    # Create a MAC of the received ciphertext
    mac = hmac.HMAC(mac_key, eval(hash_alg))
    mac.update(ciphertext)
    mac_fin = mac.finalize()
    print(f"mac_fin: {mac_fin}")
    print(f"Len mac_fin: {len(mac_fin)}")
    print(f"received_mac: {received_mac}")
    print(f"Len received_mac: {len(received_mac)}")
    if mac_fin == received_mac:
        print("Macs are the same")
        iv = ciphertext[:16]
        ciphertext = ciphertext[16:]
        print(f"LEN CIPHERTEXT CLEAR: {len(ciphertext)}")


        cipher = Cipher(algorithms.AES(encryption_key), modes.CFB(iv))

        decryptor = cipher.decryptor()
        plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        return plaintext.decode('utf-8')
    else:
        print("MACs are not the same")
        return "Error: Hashes do not match"


